@@ -712,8 +712,13 @@ void netvsc_device_remove(struct hv_device *device)
 	for (i = 0; i < net_device->num_chn; i++) {
 		/* See also vmbus_reset_channel_cb(). */
 		/* only disable enabled NAPI channel */
-		if (i < ndev->real_num_rx_queues)
+		if (i < ndev->real_num_rx_queues) {
+			netif_queue_set_napi(ndev, i, NETDEV_QUEUE_TYPE_TX,
+					     NULL);
+			netif_queue_set_napi(ndev, i, NETDEV_QUEUE_TYPE_RX,
+					     NULL);
 			napi_disable(&net_device->chan_table[i].napi);
+		}
 
 		netif_napi_del(&net_device->chan_table[i].napi);
 	}
@@ -857,7 +862,7 @@ static void netvsc_send_completion(struct net_device *ndev,
 				   msglen);
 			return;
 		}
-		fallthrough;
+		break;
 
 	case NVSP_MSG1_TYPE_SEND_RECV_BUF_COMPLETE:
 		if (msglen < sizeof(struct nvsp_message_header) +
@@ -866,7 +871,7 @@ static void netvsc_send_completion(struct net_device *ndev,
 				   msglen);
 			return;
 		}
-		fallthrough;
+		break;
 
 	case NVSP_MSG1_TYPE_SEND_SEND_BUF_COMPLETE:
 		if (msglen < sizeof(struct nvsp_message_header) +
@@ -875,7 +880,7 @@ static void netvsc_send_completion(struct net_device *ndev,
 				   msglen);
 			return;
 		}
-		fallthrough;
+		break;
 
 	case NVSP_MSG5_TYPE_SUBCHANNEL:
 		if (msglen < sizeof(struct nvsp_message_header) +
@@ -884,10 +889,6 @@ static void netvsc_send_completion(struct net_device *ndev,
 				   msglen);
 			return;
 		}
-		/* Copy the response back */
-		memcpy(&net_device->channel_init_pkt, nvsp_packet,
-		       sizeof(struct nvsp_message));
-		complete(&net_device->channel_init_wait);
 		break;
 
 	case NVSP_MSG1_TYPE_SEND_RNDIS_PKT_COMPLETE:
@@ -910,13 +911,19 @@ static void netvsc_send_completion(struct net_device *ndev,
 
 		netvsc_send_tx_complete(ndev, net_device, incoming_channel,
 					desc, budget);
-		break;
+		return;
 
 	default:
 		netdev_err(ndev,
 			   "Unknown send completion type %d received!!\n",
 			   nvsp_packet->hdr.msg_type);
+		return;
 	}
+
+	/* Copy the response back */
+	memcpy(&net_device->channel_init_pkt, nvsp_packet,
+	       sizeof(struct nvsp_message));
+	complete(&net_device->channel_init_wait);
 }
 
 static u32 netvsc_get_next_send_section(struct netvsc_device *net_device)
@@ -945,7 +952,8 @@ static void netvsc_copy_to_send_buf(struct netvsc_device *net_device,
 		     + pend_size;
 	int i;
 	u32 padding = 0;
-	u32 page_count = packet->cp_partial ? 1 : packet->page_buf_cnt;
+	u32 page_count = packet->cp_partial ? packet->rmsg_pgcnt :
+		packet->page_buf_cnt;
 	u32 remain;
 
 	/* Add padding */
@@ -1046,42 +1054,6 @@ static int netvsc_dma_map(struct hv_device *hv_dev,
 	return 0;
 }
 
-/* Build an "array" of mpb entries describing the data to be transferred
- * over VMBus. After the desc header fields, each "array" entry is variable
- * size, and each entry starts after the end of the previous entry. The
- * "offset" and "len" fields for each entry imply the size of the entry.
- *
- * The pfns are in HV_HYP_PAGE_SIZE, because all communication with Hyper-V
- * uses that granularity, even if the system page size of the guest is larger.
- * Each entry in the input "pb" array must describe a contiguous range of
- * guest physical memory so that the pfns are sequential if the range crosses
- * a page boundary. The offset field must be < HV_HYP_PAGE_SIZE.
- */
-static inline void netvsc_build_mpb_array(struct hv_page_buffer *pb,
-				u32 page_buffer_count,
-				struct vmbus_packet_mpb_array *desc,
-				u32 *desc_size)
-{
-	struct hv_mpb_array *mpb_entry = &desc->range;
-	int i, j;
-
-	for (i = 0; i < page_buffer_count; i++) {
-		u32 offset = pb[i].offset;
-		u32 len = pb[i].len;
-
-		mpb_entry->offset = offset;
-		mpb_entry->len = len;
-
-		for (j = 0; j < HVPFN_UP(offset + len); j++)
-			mpb_entry->pfn_array[j] = pb[i].pfn + j;
-
-		mpb_entry = (struct hv_mpb_array *)&mpb_entry->pfn_array[j];
-	}
-
-	desc->rangecount = page_buffer_count;
-	*desc_size = (char *)mpb_entry - (char *)desc;
-}
-
 static inline int netvsc_send_pkt(
 	struct hv_device *device,
 	struct hv_netvsc_packet *packet,
@@ -1124,11 +1096,8 @@ static inline int netvsc_send_pkt(
 
 	packet->dma_range = NULL;
 	if (packet->page_buf_cnt) {
-		struct vmbus_channel_packet_page_buffer desc;
-		u32 desc_size;
-
 		if (packet->cp_partial)
-			pb++;
+			pb += packet->rmsg_pgcnt;
 
 		ret = netvsc_dma_map(ndev_ctx->device_ctx, packet, pb);
 		if (ret) {
@@ -1136,12 +1105,11 @@ static inline int netvsc_send_pkt(
 			goto exit;
 		}
 
-		netvsc_build_mpb_array(pb, packet->page_buf_cnt,
-				(struct vmbus_packet_mpb_array *)&desc,
-				 &desc_size);
-		ret = vmbus_sendpacket_mpb_desc(out_channel,
-				(struct vmbus_packet_mpb_array *)&desc,
-				desc_size, &nvmsg, sizeof(nvmsg), req_id);
+		ret = vmbus_sendpacket_pagebuffer(out_channel,
+						  pb, packet->page_buf_cnt,
+						  &nvmsg, sizeof(nvmsg),
+						  req_id);
+
 		if (ret)
 			netvsc_dma_unmap(ndev_ctx->device_ctx, packet);
 	} else {
@@ -1290,7 +1258,7 @@ int netvsc_send(struct net_device *ndev,
 		packet->send_buf_index = section_index;
 
 		if (packet->cp_partial) {
-			packet->page_buf_cnt--;
+			packet->page_buf_cnt -= packet->rmsg_pgcnt;
 			packet->total_data_buflen = msd_len + packet->rmsg_size;
 		} else {
 			packet->page_buf_cnt = 0;
@@ -1824,6 +1792,10 @@ struct netvsc_device *netvsc_device_add(struct hv_device *device,
 	netdev_dbg(ndev, "hv_netvsc channel opened successfully\n");
 
 	napi_enable(&net_device->chan_table[0].napi);
+	netif_queue_set_napi(ndev, 0, NETDEV_QUEUE_TYPE_RX,
+			     &net_device->chan_table[0].napi);
+	netif_queue_set_napi(ndev, 0, NETDEV_QUEUE_TYPE_TX,
+			     &net_device->chan_table[0].napi);
 
 	/* Connect with the NetVsp */
 	ret = netvsc_connect_vsp(device, net_device, device_info);
@@ -1842,6 +1814,8 @@ struct netvsc_device *netvsc_device_add(struct hv_device *device,
 
 close:
 	RCU_INIT_POINTER(net_device_ctx->nvdev, NULL);
+	netif_queue_set_napi(ndev, 0, NETDEV_QUEUE_TYPE_TX, NULL);
+	netif_queue_set_napi(ndev, 0, NETDEV_QUEUE_TYPE_RX, NULL);
 	napi_disable(&net_device->chan_table[0].napi);
 
 	/* Now, we can close the channel safely */
